package org.nakedobjects.nof.reflect.peer;

import java.util.StringTokenizer;

import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.Assert;


public class MemberIdentifierImpl implements MemberIdentifier {

    private final String className;
    private final String name;
    private final String[] parameters;
    private final boolean isField;

    public MemberIdentifierImpl(final String className) {
        this.className = className;
        name = "";
        parameters = new String[0];
        isField = false;
    }

    public MemberIdentifierImpl(final String className, final String fieldName) {
        this.className = className;
        name = fieldName;
        parameters = new String[0];
        isField = true;
    }

    public MemberIdentifierImpl(final String className, final String methodName, final Class[] parameterClasses) {
        this.className = className;
        name = methodName;
        parameters = new String[parameterClasses == null ? 0 : parameterClasses.length];
        for (int i = 0; i < parameters.length; i++) {
            parameters[i] = parameterClasses[i].getName();
        }
        isField = false;
    }

    public MemberIdentifierImpl(final String className, final String methodName, final String[] parameterClasses) {
        this.className = className;
        name = methodName;
        parameters = new String[parameterClasses == null ? 0 : parameterClasses.length];
        for (int i = 0; i < parameters.length; i++) {
            parameters[i] = parameterClasses[i];
        }
        isField = false;
    }

    public MemberIdentifierImpl(final String className, final String methodName, final NakedObjectSpecification[] specifications) {
        this.className = className;
        name = methodName;
        parameters = new String[specifications == null ? 0 : specifications.length];
        for (int i = 0; i < parameters.length; i++) {
            parameters[i] = specifications[i].getFullName();
        }
        isField = false;
    }

    public boolean equals(final Object obj) {
        if (this == obj) {
            return true;
        } else if (obj instanceof MemberIdentifierImpl) {
            MemberIdentifierImpl other = (MemberIdentifierImpl) obj;
            return equals(other.className, className) && equals(other.name, other.name) && equals(other.parameters, parameters);
        }
        return false;
    }

    private boolean equals(final String[] a, final String[] b) {
        if (a == null && b == null) {
            return true;
        }

        if (a == null && b != null) {
            return false;
        }

        if (a != null && b == null) {
            return false;
        }

        if (a.length != b.length) {
            return false;
        }

        for (int i = 0; i < b.length; i++) {
            if (!a[i].equals(b[i])) {
                return false;
            }
        }

        return true;
    }

    private boolean equals(final String a, final String b) {
        if (a == b) {
            return true;
        }

        if (a != null) {
            return a.equals(b);
        }

        return false;
    }

    public String getClassName() {
        return className;
    }

    public String getName() {
        return name;
    }

    public String[] getParameters() {
        return parameters;
    }

    public boolean isField() {
        return isField;
    }

    private String asString = null;

    public String toString() {
        if (asString == null) {
            StringBuffer str = new StringBuffer();
            str.append(className);
            str.append('#');
            str.append(name);
            str.append('(');
            for (int i = 0; i < parameters.length; i++) {
                if (i > 0) {
                    str.append(", ");
                }
                str.append(parameters[i]);
            }
            str.append(')');
            asString = str.toString();
        }
        return asString;
    }

    public String toIdentityString(final int depth) {
        Assert.assertTrue(depth >= MemberIdentifier.CLASS && depth <= MemberIdentifier.PARMS);
        switch (depth) {
        case (MemberIdentifier.CLASS):
            return toClassIdentityString();
        case (MemberIdentifier.CLASS_NAME):
            return toClassAndNameIdentityString();
        case (MemberIdentifier.CLASS_NAME_PARMS):
            return toFullIdentityString();
        case (MemberIdentifier.NAME):
            return toNameIdentityString();
        case (MemberIdentifier.PARMS):
            return toParmsIdentityString();
        }
        return null;
    }

    public String toClassIdentityString() {
        return className;
    }

    public String toNameIdentityString() {
        return name;
    }

    public String toClassAndNameIdentityString() {
        return toClassIdentityString() + "#" + name;
    }

    public String toParmsIdentityString() {
        final StringBuffer str = new StringBuffer();
        if (!isField()) {
            str.append('(');
            for (int i = 0; i < parameters.length; i++) {
                if (i > 0) {
                    str.append(",");
                }
                str.append(parameters[i]);
            }
            str.append(')');
        }
        return str.toString();
    }

    private String identityString;

    public String toFullIdentityString() {
        if (identityString == null) {
            if (name.length() == 0) {
                identityString = toClassIdentityString();
            } else {
                final StringBuffer str = new StringBuffer();
                str.append(toClassAndNameIdentityString());
                str.append(toParmsIdentityString());
                identityString = str.toString();
            }
        }
        return identityString;
    }

    public static MemberIdentifierImpl fromIdentityString(final String asString) {
        Assert.assertNotNull(asString);
        final int indexOfHash = asString.indexOf("#");
        final int indexOfOpenBracket = asString.indexOf("(");
        final int indexOfCloseBracket = asString.indexOf(")");
        final String className = asString.substring(0, indexOfHash == -1 ? asString.length() : indexOfHash);
        if (indexOfHash == -1 || indexOfHash == (asString.length() - 1)) {
            return new MemberIdentifierImpl(className);
        }
        String name = null;
        if (indexOfOpenBracket == -1) {
            name = asString.substring(indexOfHash + 1);
            return new MemberIdentifierImpl(className, name);
        }
        String[] parms = null;
        name = asString.substring(indexOfHash + 1, indexOfOpenBracket);
        final String allParms = asString.substring(indexOfOpenBracket + 1, indexOfCloseBracket).trim();
        if (allParms.length() > 0) {
            // use stringtokenizer for dotnet compatibility
            final StringTokenizer tokens = new StringTokenizer(allParms, ",", false);
            parms = new String[tokens.countTokens()];
            for (int i = 0; tokens.hasMoreTokens(); i++) {
                parms[i] = tokens.nextToken();
            }
        }
        return new MemberIdentifierImpl(className, name, parms);
    }

    public int compareTo(Object o2) {
        return toString().compareTo(o2.toString());
    }

    public NakedObjectSpecification[] getParameterSpecifications() {
        final NakedObjectSpecification[] specifications = new NakedObjectSpecification[parameters.length];
        for (int i = 0; i < parameters.length; i++) {
            specifications[i] = NakedObjectsContext.getReflector().loadSpecification(parameters[i]);
        }
        return specifications;
    }

}
// Copyright (c) Naked Objects Group Ltd.
