package org.nakedobjects.nof.reflect.remote;

import org.apache.log4j.Category;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.reflect.peer.AbstractOneToManyPeer;
import org.nakedobjects.nof.reflect.peer.OneToManyPeer;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.IdentityData;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;


/**
 * A reflection peer for changing one-to-many fields remotely, instead of on the local machine. Any requests
 * to add or remove elements from the field will be passed over the network to the server for completion. Only
 * requests on persistent objects are passed to the server; on a transient object the request will always be
 * dealt with locally.
 * 
 * <p>
 * If any of the objects involved have been changed on the server by another process then a
 * ConcurrencyException will be passed back to the client and re-thrown.
 * </p>
 */
final class ProxyOneToManyAssociation extends AbstractOneToManyPeer {
    private final static Category LOG = Category.getInstance(ProxyOneToManyAssociation.class);
    private Distribution connection;
    private final ObjectEncoder encoder;

    public ProxyOneToManyAssociation(
            final OneToManyPeer local,
            final Distribution connection,
            final ObjectEncoder encoder) {
        super(local);
        this.connection = connection;
        this.encoder = encoder;
    }

    public void addAssociation(final NakedObject inObject, final NakedObject associate) {
        if (executeRemotely(inObject)) {
            LOG.debug("set association remotely " + getIdentifier() + " in " + inObject + " with " + associate);
            try {
                IdentityData targetReference = encoder.createIdentityData(inObject);
                IdentityData associateReference = encoder.createIdentityData(associate);
                ObjectData[] updates;
                updates = connection.setAssociation(NakedObjectsContext.getSession(), getIdentifier().getName(), targetReference,
                        associateReference);
                updateChangedObjects(updates);
            } catch (ConcurrencyException e) {
                throw concurrencyException(e);
            } catch (NakedObjectRuntimeException e) {
                LOG.error("remote exception: " + e.getMessage(), e);
                throw e;
            }
        } else {
            LOG.debug("set association locally " + getIdentifier() + " in " + inObject + " with " + associate);
            super.addAssociation(inObject, associate);
        }
    }

    private ConcurrencyException concurrencyException(final ConcurrencyException e) {
        LOG.info("concurrency conflict: " + e.getMessage());
        Oid source = e.getSource();
        if (source == null) {
            return e;
        } else {
            NakedObject failedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(source);
            NakedObjectsContext.getObjectPersistor().reload(failedObject);
            return new ConcurrencyException("Object automatically reloaded: " + failedObject.titleString(), e);
        }
    }

    private void updateChangedObjects(final ObjectData[] updates) {
        for (int i = 0; i < updates.length; i++) {
            LOG.debug("update " + updates[i].getOid());
            encoder.restore(updates[i]);
        }
    }

    private boolean executeRemotely(final NakedObject target) {
        return target.getResolveState().isPersistent();
    }

    /**
     * Remove an associated object (the element) from the specified NakedObject in the association field
     * represented by this object.
     */
    public void removeAssociation(final NakedObject inObject, final NakedObject associate) {
        if (executeRemotely(inObject)) {
            LOG.debug("clear association remotely " + inObject + "/" + associate);
            try {
                IdentityData targetReference = encoder.createIdentityData(inObject);
                IdentityData associateReference = encoder.createIdentityData(associate);
                ObjectData[] updates;
                updates = connection.clearAssociation(NakedObjectsContext.getSession(), getIdentifier().getName(),
                        targetReference, associateReference);
                updateChangedObjects(updates);
            } catch (ConcurrencyException e) {
                throw concurrencyException(e);
            } catch (NakedObjectRuntimeException e) {
                LOG.error("remote exception: " + e.getMessage(), e);
                throw e;
            }
        } else {
            LOG.debug("clear association locally " + inObject + "/" + associate);
            super.removeAssociation(inObject, associate);
        }
    }
}
// Copyright (c) Naked Objects Group Ltd.
