package org.nakedobjects.nof.reflect.remote;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedReference;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.adapter.Version;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.noa.persist.InstancesCriteria;
import org.nakedobjects.noa.persist.ObjectPersistenceException;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.persist.AbstractObjectPersistor;
import org.nakedobjects.nof.core.persist.TransactionException;
import org.nakedobjects.nof.core.util.DebugString;
import org.nakedobjects.nof.core.util.NotImplementedException;
import org.nakedobjects.nof.reflect.remote.data.ClientActionResultData;
import org.nakedobjects.nof.reflect.remote.data.ClientSideTransaction;
import org.nakedobjects.nof.reflect.remote.data.ClientTransactionEvent;
import org.nakedobjects.nof.reflect.remote.data.CriteriaData;
import org.nakedobjects.nof.reflect.remote.data.Data;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.IdentityData;
import org.nakedobjects.nof.reflect.remote.data.KnownObjects;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;
import org.nakedobjects.nof.reflect.remote.data.ReferenceData;


public class ProxyPersistor extends AbstractObjectPersistor {
    final static Logger LOG = Logger.getLogger(ProxyPersistor.class);
    private Distribution connection;
    private ObjectEncoder encoder;
    private ClientSideTransaction clientSideTransaction;
    private boolean checkObjectsForDirtyFlag;

    private Hashtable cache = new Hashtable();

    public void abortTransaction() {
        checkTransactionInProgress();
        LOG.debug("abortTransaction");
        clientSideTransaction.rollback();
        clientSideTransaction = null;
    }

    public synchronized void destroyObject(NakedObject object) {
        checkTransactionInProgress();
        LOG.debug("destroyObject " + object);
        clientSideTransaction.addDestroyObject(object);

        // TODO need to do garbage collection instead
        // NakedObjects.getObjectLoader().unloaded(object);
    }

    private void checkTransactionInProgress() {
        if (clientSideTransaction == null) {
            throw new TransactionException("No transaction in progress");
        }
    }

    public boolean flushTransaction() {
        return false;
    }

    public void endTransaction() {
        checkTransactionInProgress();
        LOG.debug("endTransaction");

        if (clientSideTransaction.isEmpty()) {
            LOG.debug("  no transaction commands to process");
            clientSideTransaction = null;
            return;
        }

        KnownObjects knownObjects = new KnownObjects();

        ClientTransactionEvent[] entries = clientSideTransaction.getEntries();
        int length = entries.length;
        int[] types = new int[length];
        ReferenceData[] data = new ReferenceData[length];
        for (int i = 0; i < length; i++) {
            types[i] = entries[i].getType();
            switch (types[i]) {
            case ClientTransactionEvent.ADD:
                data[i] = encoder.createMakePersistentGraph(entries[i].getObject(), knownObjects);
                break;
            case ClientTransactionEvent.CHANGE:
                data[i] = encoder.createGraphForChangedObject(entries[i].getObject(), knownObjects);
                break;
            case ClientTransactionEvent.DELETE:
                data[i] = encoder.createIdentityData(entries[i].getObject());
                break;
            }
        }

        ClientActionResultData results;
        try {
            results = connection.executeClientAction(NakedObjectsContext.getSession(), data, types);
        } catch (ConcurrencyException e) {
            LOG.info("concurrency conflict: " + e.getMessage());
            Oid oid = e.getSource();
            if (oid == null) {
                throw e;
            } else {
                NakedObject failedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(oid);
                reload(failedObject);
                throw new ConcurrencyException("Object automatically reloaded: " + failedObject.titleString(), e);
            }
        }

        if (results != null) {
            ReferenceData[] persistedUpdates = results.getPersisted();
            Version[] changedVersions = results.getChanged();
            for (int i = 0; i < length; i++) {
                switch (types[i]) {
                case ClientTransactionEvent.ADD:
                    ReferenceData update = persistedUpdates[i];
                    NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(update.getOid()); // causes OID to be updated
                    adapter.changeState(ResolveState.RESOLVED);
                    entries[i].getObject().setOptimisticLock(update.getVersion());
                    
                    break;
                case ClientTransactionEvent.CHANGE:
                    entries[i].getObject().setOptimisticLock(changedVersions[i]);
                    NakedObjectsContext.getUpdateNotifer().addChangedObject(entries[i].getObject());
                    break;
                }
            }
            
            ObjectData[] updates = results.getUpdates();
            for (int i = 0; i < updates.length; i++) {
                LOG.debug("update " + updates[i].getOid());
                encoder.restore(updates[i]);
            }
            
            for (int i = 0; i < length; i++) {
                switch (types[i]) {
                case ClientTransactionEvent.DELETE:
                    NakedObjectsContext.getUpdateNotifer().addDisposedObject(entries[i].getObject());
                    break;
                }
            }
        }

        clientSideTransaction = null;
    }

    public void debugData(final DebugString debug) {
        super.debugData(debug);
        debug.appendln("Connection", connection);
    }

    protected ClientSideTransaction getClientSideTransaction() {
        return clientSideTransaction;
    }

    public String debugTitle() {
        return "Proxy Object Manager";
    }

    protected NakedObject[] getInstances(final InstancesCriteria criteria) {
        NakedObjectSpecification specification = criteria.getSpecification();
        LOG.debug("getInstances of " + specification + " with " + criteria);
        if (cache.containsKey(specification)) {
            Object collection = cache.get(specification);
            if (collection instanceof NakedCollection) {
                NakedCollection c = (NakedCollection) collection;
                Vector instances = new Vector();
                Enumeration e = c.elements();
                while (e.hasMoreElements()) {
                    NakedObject instance = (NakedObject) e.nextElement();
                    if (criteria.matches(instance)) {
                        instances.addElement(instance);
                    }
                }
                return (NakedObject[]) instances.toArray(new NakedObject[instances.size()]);
            }
        }

        CriteriaData criteriaData = encoder.createCriteriaData(criteria);
        ObjectData[] instances = connection.findInstances(NakedObjectsContext.getSession(), criteriaData);
        NakedObject[] instances1 = new NakedObject[instances.length];
        for (int i = 0; i < instances.length; i++) {
            instances1[i] = (NakedObject) encoder.restore(instances[i]);
        }
        clearChanges();
        return instances1;
    }

    public synchronized NakedObject getObject(final Oid oid, final NakedObjectSpecification hint) {
        NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(oid);
        if (adapter != null) { 
            return adapter;
        } else {
            ObjectData data = connection.getObject(null, oid, hint.getFullName());
            return (NakedObject) encoder.restore(data);
        }
    }

    public boolean hasInstances(final NakedObjectSpecification specification, final boolean includeSubclasses) {
        LOG.debug("hasInstances of " + specification);
        if (cache.containsKey(specification)) {
            Object collection = cache.get(specification);
            if (collection instanceof NakedCollection) {
                return ((NakedCollection) collection).size() > 0;
            }
        }

        // TODO includeInstances flag is not passed to server
        return connection.hasInstances(NakedObjectsContext.getSession(), specification.getFullName());
    }

    public boolean isInitialized() {
        return true;
    }

    public void init() {
        connection.init();
    }

    public synchronized void makePersistent(NakedObject object) {
        checkTransactionInProgress();
        LOG.debug("makePersistent " + object);
        clientSideTransaction.addMakePersistent(object);
    }

    public void objectChanged(final NakedObject object) {
        if (object.getResolveState().isTransient()) {
            NakedObjectsContext.getUpdateNotifer().addChangedObject(object);
        }

        if (object.getResolveState().respondToChangesInPersistentObjects()) {
            checkTransactionInProgress();
            clientSideTransaction.addObjectChanged(object);
        }
    }

    public void reset() {}

    public void reload(final NakedObject object) {
        ObjectData update = connection.resolveImmediately(NakedObjectsContext.getSession(), encoder.createIdentityData(object));
        encoder.restore(update);
    }

    public synchronized void resolveImmediately(NakedObject object) {
        ResolveState resolveState = object.getResolveState();
        if (resolveState.isResolvable(ResolveState.RESOLVING)) {
            Oid oid = object.getOid();
            LOG.debug("resolve object (remotely from server)" + oid);
            ObjectData data = connection.resolveImmediately(NakedObjectsContext.getSession(), encoder.createIdentityData(object));
            encoder.restore(data);
        }
    }

    public void resolveField(final NakedObject object, final NakedObjectField field) {
        if (field.isValue()) {
            return;
        }
        NakedReference reference = (NakedReference) field.get(object);
        if (reference != null && reference.getResolveState().isResolved()) {
            return;
        }
        if (reference == null || !reference.getResolveState().isPersistent()) {
            return;
        }

        LOG.info("resolve-eagerly on server " + object + "/" + field.getId());
        Data data = connection.resolveField(NakedObjectsContext.getSession(), encoder.createIdentityData(object), field.getId());
        encoder.restore(data);
    }

    public void saveChanges() {
        if (checkObjectsForDirtyFlag) {
            LOG.debug("collating changed objects");
            Enumeration e = NakedObjectsContext.getObjectLoader().getIdentifiedObjects();
            while (e.hasMoreElements()) {
                Object o = e.nextElement();
                if (o instanceof NakedObject) {
                    NakedObject object = (NakedObject) o;
                    if (object.getSpecification().isDirty(object)) {
                        LOG.debug("  found dirty object " + object);
                        objectChanged(object);
                        object.getSpecification().clearDirty(object);
                    }
                }
            }
        }
    }

    private synchronized void clearChanges() {
        if (checkObjectsForDirtyFlag) {
            LOG.debug("clearing changed objects");
            Enumeration e = NakedObjectsContext.getObjectLoader().getIdentifiedObjects();
            while (e.hasMoreElements()) {
                Object o = e.nextElement();
                if (o instanceof NakedObject) {
                    NakedObject object = (NakedObject) o;
                    if (object.getSpecification().isDirty(object)) {
                        LOG.debug("  found dirty object " + object);
                        object.getSpecification().clearDirty(object);
                    }
                }
            }
        }
    }

    /**
     * .NET property
     * 
     * @property
     */
    public void set_Connection(final Distribution connection) {
        this.connection = connection;
    }

    /**
     * .NET property
     * 
     * @property
     */
    public void set_Encoder(final ObjectEncoder factory) {
        this.encoder = factory;
    }

    public void setConnection(final Distribution connection) {
        this.connection = connection;
    }

    public void setEncoder(final ObjectEncoder factory) {
        this.encoder = factory;
    }

    public void startTransaction() {
        LOG.debug("startTransaction");
        clearChanges();
        if (clientSideTransaction == null) {
            clientSideTransaction = new ClientSideTransaction();
        } else {
            ClientSideTransaction transaction = clientSideTransaction;
            clientSideTransaction = null;
            throw new ObjectPersistenceException("Can't start transaction when one already started: " + transaction);
        }
    }

    public void shutdown() {
        connection.shutdown();
    }

    /**
     * Expose as a .NET property
     * 
     * @property
     */
    public void set_CheckObjectsForDirtyFlag(final boolean checkObjectsForDirtyFlag) {
        this.checkObjectsForDirtyFlag = checkObjectsForDirtyFlag;
    }

    public void setCheckObjectsForDirtyFlag(final boolean checkObjectsForDirtyFlag) {
        this.checkObjectsForDirtyFlag = checkObjectsForDirtyFlag;
    }

    /**
     * Expose as a .NET property
     * 
     * @property
     */
    public void set_CacheInstances(final String[] names) {
        setCacheInstances(names);
    }

    public void setCacheInstances(final String[] names) {
        for (int i = 0; i < names.length; i++) {
            NakedObjectSpecification spec = NakedObjectsContext.getReflector().loadSpecification(names[i]);
            cache.put(spec, Boolean.TRUE);
        }
    }

    public Oid getOidForService(String name) {
        IdentityData data = connection.oidForService(NakedObjectsContext.getSession(), name);
        return data.getOid();
    }
    
    public void setServices(Object[] services) {}
    
    public void registerService(String name, Oid oid) {
        throw new NotImplementedException();
    }
}
// Copyright (c) Naked Objects Group Ltd.
