package org.nakedobjects.nof.reflect.remote;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.persist.ConcurrencyException;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.reflect.peer.AbstractValuePeer;
import org.nakedobjects.nof.reflect.peer.ValuePeer;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.IdentityData;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;
import org.nakedobjects.nof.reflect.remote.data.ValueData;


/**
 * A reflection peer for changing value fields remotely, instead of on the local machine. Any requests to set
 * or clear a field will be passed over the network to the server for completion. Only requests on persistent
 * objects are passed to the server; on a transient object the request will always be dealt with locally.
 * 
 * <p>
 * If the target object has been changed on the server by another process then a ConcurrencyException will be
 * passed back to the client and re-thrown.
 * </p>
 */
final class ProxyValueAssociation extends AbstractValuePeer {
    private final static Logger LOG = Logger.getLogger(ProxyValueAssociation.class);
    private final Distribution connection;
    private final ObjectEncoder encoder;

    public ProxyValueAssociation(final ValuePeer local, final Distribution connection, final ObjectEncoder encoder) {
        super(local);
        this.connection = connection;
        this.encoder = encoder;
    }

    private ConcurrencyException concurrencyException(final ConcurrencyException e) {
        LOG.info("concurrency conflict: " + e.getMessage());
        Oid source = e.getSource();
        if (source == null) {
            return e;
        } else {
            NakedObject failedObject = NakedObjectsContext.getObjectLoader().getAdapterFor(source);
            NakedObjectsContext.getObjectPersistor().reload(failedObject);
            return new ConcurrencyException("Object automatically reloaded: " + failedObject.titleString(), e);
        }
    }

    private boolean executeRemotely(final NakedObject target) {
        return target.getResolveState().isPersistent();
    }

    public void setValue(final NakedObject inObject, final NakedValue value) {
        if (executeRemotely(inObject)) {
            LOG.debug("set value remotely " + getIdentifier() + " in " + inObject + " with " + value);
            IdentityData targetReference = encoder.createIdentityData(inObject);
            ObjectData[] updates;
            try {
                Session session = NakedObjectsContext.getSession();
                String fieldName = getIdentifier().getName();
                ValueData val = encoder.createValue(value);
                updates = connection.setValue(session, fieldName, targetReference, val);
            } catch (ConcurrencyException e) {
                throw concurrencyException(e);
            }
            updateChangedObjects(updates);
        } else {
            LOG.debug("set value locally " + getIdentifier() + " in " + inObject + " with " + value);
            super.setValue(inObject, value);
        }
    }

    private void updateChangedObjects(final ObjectData[] updates) {
        for (int i = 0; i < updates.length; i++) {
            LOG.debug("update " + updates[i].getOid());
            encoder.restore(updates[i]);
        }
    }
    
}
// Copyright (c) Naked Objects Group Ltd.
