package org.nakedobjects.nof.reflect.remote;

import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.noa.adapter.Naked;
import org.nakedobjects.noa.adapter.NakedCollection;
import org.nakedobjects.noa.adapter.NakedObject;
import org.nakedobjects.noa.adapter.NakedValue;
import org.nakedobjects.noa.adapter.Oid;
import org.nakedobjects.noa.adapter.ResolveState;
import org.nakedobjects.noa.adapter.Version;
import org.nakedobjects.noa.persist.InstancesCriteria;
import org.nakedobjects.noa.persist.NakedObjectPersistor;
import org.nakedobjects.noa.reflect.NakedObjectAction;
import org.nakedobjects.noa.reflect.NakedObjectField;
import org.nakedobjects.noa.reflect.NakedObjectMember;
import org.nakedobjects.noa.reflect.OneToManyAssociation;
import org.nakedobjects.noa.reflect.OneToOneAssociation;
import org.nakedobjects.noa.reflect.ValueAssociation;
import org.nakedobjects.noa.security.Session;
import org.nakedobjects.noa.security.SessionManager;
import org.nakedobjects.noa.spec.NakedObjectSpecification;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.security.PasswordAuthenticationRequest;
import org.nakedobjects.nof.core.service.ServiceUtil;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.core.util.NakedObjectConfiguration;
import org.nakedobjects.nof.core.util.UnknownTypeException;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.MemberIdentifierImpl;
import org.nakedobjects.nof.reflect.remote.data.ClientActionResultData;
import org.nakedobjects.nof.reflect.remote.data.ClientTransactionEvent;
import org.nakedobjects.nof.reflect.remote.data.CriteriaData;
import org.nakedobjects.nof.reflect.remote.data.Data;
import org.nakedobjects.nof.reflect.remote.data.Distribution;
import org.nakedobjects.nof.reflect.remote.data.IdentityData;
import org.nakedobjects.nof.reflect.remote.data.KnownObjects;
import org.nakedobjects.nof.reflect.remote.data.NullData;
import org.nakedobjects.nof.reflect.remote.data.ObjectData;
import org.nakedobjects.nof.reflect.remote.data.ObjectEncoder;
import org.nakedobjects.nof.reflect.remote.data.ReferenceData;
import org.nakedobjects.nof.reflect.remote.data.ServerActionResultData;
import org.nakedobjects.nof.reflect.remote.data.ValueData;
import org.nakedobjects.nof.reflect.spec.ActionImpl;
import org.nakedobjects.object.RequiresSetup;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Properties;

import org.apache.log4j.Logger;


public class ServerDistribution implements RequiresSetup, Distribution {
    private static final Logger LOG = Logger.getLogger(ServerDistribution.class);
    private ObjectEncoder encoder;
    private final SessionManager sessionManager;

    public ServerDistribution(SessionManager sessionManager) {
        this.sessionManager = sessionManager;
    }

    public Session authenticate(String userNameAndPassword) {
        PasswordAuthenticationRequest request = extractAuthenticationRequest(userNameAndPassword);
        return sessionManager.authenticate(request);
    }

    private PasswordAuthenticationRequest extractAuthenticationRequest(String userNameAndPassword) {
        int delimiter = userNameAndPassword.indexOf("::");
        String username = userNameAndPassword.substring(0, delimiter);
        String password = userNameAndPassword.substring(delimiter + 2);
        PasswordAuthenticationRequest request = new PasswordAuthenticationRequest(username, password);
        return request;
    }

    public boolean authoriseVisibility(final Session session, final String memberName) {
        return getMember(memberName).isVisibleForSession();
    }

    public boolean authoriseUsability(final Session session, final String memberName) {
        return getMember(memberName).isUsableForSession().isAllowed();
    }

    private NakedObjectMember getMember(final String memberName) {
        final MemberIdentifier id = MemberIdentifierImpl.fromIdentityString(memberName);
        final NakedObjectSpecification specification = NakedObjectsContext.getReflector().loadSpecification(id.getClassName());
        NakedObjectMember member;
        if (id.isField()) {
            member = (NakedObjectMember) specification.getField(id.getName());
            if (member == null) {
                throw new IllegalRequestException("No field found for id " + id);
            }
        } else {
            member = (NakedObjectMember) specification.getObjectAction(NakedObjectAction.USER, id.getName(), id
                    .getParameterSpecifications());
            if (member == null) {
                throw new IllegalRequestException("No user action found for id " + id);
            }
        }
        return member;
    }

    public ObjectData[] clearAssociation(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final IdentityData associated) {
        LOG.debug("request clearAssociation " + fieldIdentifier + " on " + target + " of " + associated + " for " + session);
        NakedObject inObject = getPersistentNakedObject(session, target);
        NakedObject associate = getPersistentNakedObject(session, associated);
        NakedObjectSpecification specification = inObject.getSpecification();
        NakedObjectField association = (NakedObjectField) specification.getField(fieldIdentifier);
        if (!association.isVisible() || association.isUsable(inObject).isVetoed()) {
            throw new IllegalRequestException("can't modify field as not visible or editable");
        }
        if (association instanceof OneToOneAssociation) {
            ((OneToOneAssociation) association).clearAssociation(inObject, associate);
        } else {
            ((OneToManyAssociation) association).removeElement(inObject, associate);
        }
        return getUpdates();
    }

    public ObjectData[] clearValue(final Session session, final String fieldIdentifier, final IdentityData target) {
        LOG.debug("request clearValue " + fieldIdentifier + " on " + target + " for " + session);
        NakedObject inObject = getPersistentNakedObject(session, target);
        ValueAssociation association = (ValueAssociation) inObject.getSpecification().getField(fieldIdentifier);
        if (!association.isVisible() || association.isUsable(inObject).isVetoed()) {
            throw new IllegalRequestException("can't modify field as not visible or editable");
        }

        association.clearValue(inObject);
        return getUpdates();
    }

    public void closeSession(Session session) {
        sessionManager.closeSession(session);
    }

    private ObjectData[] convertToNakedCollection(final NakedCollection instances) {
        ObjectData[] data = new ObjectData[instances.size()];
        Enumeration elements = instances.elements();
        int i = 0;
        while (elements.hasMoreElements()) {
            NakedObject element = (NakedObject) elements.nextElement();
            data[i++] = encoder.createCompletePersistentGraph(element );
        }
        return data;
    }

    public ServerActionResultData executeServerAction(
            final Session session,
            final String actionType,
            final String actionIdentifier,
            final ReferenceData target,
            final Data[] parameterData) {
        LOG.debug("request executeAction " + actionIdentifier + " on " + target + " for " + session);

        NakedObject object;
        KnownObjects knownObjects = new KnownObjects();
        if (target instanceof IdentityData) {
            object = getPersistentNakedObject(session, (IdentityData) target);
        } else if (target instanceof ObjectData) {
            object = (NakedObject) encoder.restore(target, knownObjects);
        } else if (target == null) {
            object = null;
        } else {
            throw new NakedObjectRuntimeException();
        }

        NakedObjectAction action = getActionMethod(actionType, actionIdentifier, parameterData, object);
        Naked[] parameters = getParameters(session, parameterData, knownObjects);

        if (action == null) {
            throw new NakedObjectsRemoteException("Could not find method " + actionIdentifier);
        }

        Naked result = action.execute(object, parameters);

        ObjectData persistedTarget;
        if (target == null) {
            persistedTarget = null;
        } else if (target instanceof ObjectData) {
            persistedTarget = encoder.createMadePersistentGraph((ObjectData) target, object);
        } else {
            persistedTarget = null;
        }

        ObjectData[] persistedParameters = new ObjectData[parameterData.length];
        for (int i = 0; i < persistedParameters.length; i++) {
            if (action.getParameterTypes()[i].getType() == NakedObjectSpecification.OBJECT
                    && parameterData[i] instanceof ObjectData) {
                persistedParameters[i] = encoder.createMadePersistentGraph((ObjectData) parameterData[i],
                        (NakedObject) parameters[i]);
            }
        }
        String[] messages = NakedObjectsContext.getMessageBroker().getMessages();
        String[] warnings = NakedObjectsContext.getMessageBroker().getWarnings();

        // TODO for efficiency, need to remove the objects in the results graph from the updates set
        return encoder.createServerActionResult(result, getUpdates(), getDisposed(), persistedTarget, persistedParameters, messages, warnings);
    }

    public ClientActionResultData executeClientAction(final Session session, final ReferenceData[] data, final int[] types) {
        LOG.debug("execute client action for " + session);
        LOG.debug("start transaction");
        NakedObjectPersistor persistor = persistor();
        persistor.startTransaction();
        try {
            KnownObjects knownObjects = new KnownObjects();
            NakedObject[] persistedObjects = new NakedObject[data.length];
            NakedObject[] disposedObjects = new NakedObject[data.length];
            NakedObject[] changedObjects = new NakedObject[data.length];
            for (int i = 0; i < data.length; i++) {
                NakedObject object;
                switch (types[i]) {
                case ClientTransactionEvent.ADD:
                    object = (NakedObject) encoder.restore(data[i], knownObjects);
                    persistedObjects[i] = object;
                    if (object.getOid().isTransient()) { // confirm that the graph has not been made
                        // persistent earlier in this loop
                        LOG.debug("  makePersistent " + data[i]);
                        persistor.makePersistent(object);
                    }
                    break;

                case ClientTransactionEvent.CHANGE:
                    NakedObject obj = getPersistentNakedObject(data[i]);
                    obj.checkLock(data[i].getVersion());

                    object = (NakedObject) encoder.restore(data[i], knownObjects);
                    LOG.debug("  objectChanged " + data[i]);
                    persistor.objectChanged(object);
                    changedObjects[i] = object;
                    break;

                case ClientTransactionEvent.DELETE:
                    NakedObject inObject = getPersistentNakedObject(data[i]);
                    inObject.checkLock(data[i].getVersion());

                    LOG.debug("  destroyObject " + data[i] + " for " + session);
                    disposedObjects[i] = inObject;
                    persistor.destroyObject(inObject);
                    break;
                }

            }

            LOG.debug("  end transaction");
            persistor.endTransaction();

            ReferenceData[] madePersistent = new ReferenceData[data.length];
            Version[] changedVersion = new Version[data.length];

            for (int i = 0; i < data.length; i++) {
                switch (types[i]) {
                case ClientTransactionEvent.ADD:
                    madePersistent[i] = encoder.createIdentityData(persistedObjects[i]);
                    break;

                case ClientTransactionEvent.CHANGE:
                    changedVersion[i] = changedObjects[i].getVersion();
                    break;

                }
            }

            return encoder.createClientActionResult(madePersistent, changedVersion, getUpdates());
        } catch (RuntimeException e) {
            LOG.info("abort transaction", e);
            persistor.abortTransaction();
            throw e;
        }
    }

    public ObjectData[] findInstances(final Session session, final CriteriaData criteriaData) {
        InstancesCriteria criteria = encoder.restoreCriteria(criteriaData);
        LOG.debug("request findInstances " + criteria + " for " + session);
        NakedCollection instances = persistor().findInstances(criteria);
        return convertToNakedCollection(instances);
    }

    private NakedObjectAction getActionMethod(
            final String actionType,
            final String actionIdentifier,
            final Data[] parameterData,
            final NakedObject object) {
        NakedObjectSpecification[] parameterSpecifiactions = new NakedObjectSpecification[parameterData.length];
        for (int i = 0; i < parameterSpecifiactions.length; i++) {
            parameterSpecifiactions[i] = getSpecification(parameterData[i].getType());
        }

        NakedObjectAction.Type type = ActionImpl.getType(actionType);

        int pos = actionIdentifier.indexOf('#');
        String className = actionIdentifier.substring(0, pos);
        String methodName = actionIdentifier.substring(pos + 1);

        NakedObjectAction action;
        if (object == null) {
            action = (NakedObjectAction) NakedObjectsContext.getReflector().loadSpecification(className).getClassAction(type,
                    methodName, parameterSpecifiactions);
        } else {
            action = (NakedObjectAction) object.getSpecification().getObjectAction(type, methodName, parameterSpecifiactions);
        }
        return action;
    }

    public ObjectData getObject(Session session, Oid oid, String specificationName) {
        NakedObjectSpecification specification = getSpecification(specificationName);
        NakedObject object = persistor().getObject(oid, specification);
        return encoder.createForUpdate(object);
    }

    private Naked[] getParameters(final Session session, final Data[] parameterData, final KnownObjects knownObjects) {
        Naked[] parameters = new Naked[parameterData.length];
        for (int i = 0; i < parameters.length; i++) {
            Data data = parameterData[i];
            if (data instanceof NullData) {
                continue;
            }

            if (data instanceof IdentityData) {
                parameters[i] = getPersistentNakedObject(session, (IdentityData) data);
            } else if (data instanceof ObjectData) {
                parameters[i] = encoder.restore((ObjectData) data, knownObjects);
            } else if (data instanceof ValueData) {
                NakedObjectSpecification valueSpecification = NakedObjectsContext.getReflector()
                        .loadSpecification(data.getType());
                NakedValue value = NakedObjectsContext.getObjectLoader().createValueInstance(valueSpecification);
                String valueData = ((ValueData) data).getEncodedValue();
                value.restoreFromEncodedString(valueData);
                parameters[i] = value;
            } else {
                throw new UnknownTypeException(data);
            }
        }
        return parameters;
    }

    private NakedObject getPersistentNakedObject(final Session session, final IdentityData object) {
        NakedObject obj = getPersistentNakedObject(object);
        LOG.debug("get object " + object + " for " + session + " --> " + obj);
        obj.checkLock(object.getVersion());
        return obj;
    }

    private NakedObject getPersistentNakedObject(final ReferenceData object) {
        NakedObjectSpecification spec = getSpecification(object.getType());
        NakedObject obj = NakedObjectsContext.getObjectPersistor().getObject(object.getOid(), spec);
        Assert.assertNotNull(obj);
        return obj;
    }

    public Properties getProperties() {
        Properties properties = new Properties();
        properties.put("test-client", "true");

        // pass over services
        NakedObjectConfiguration configuration = NakedObjectsContext.getConfiguration();
        NakedObjectConfiguration serviceProperties = configuration.getProperties(Configuration.ROOT + "services");
        Enumeration e = serviceProperties.propertyNames();
        while (e.hasMoreElements()) {
            String name = (String) e.nextElement();
            properties.put(name, serviceProperties.getString(name));
        }

        // pass over OID generator
        String oidGeneratorClass = configuration.getString("oid-generator");
        // TODO use persistor to find out this name, eg
        // NakedObjectsContext.getObjectPersistor().getOidGenerator()
        if (oidGeneratorClass != null) {
            properties.put(Configuration.ROOT + "persistor.oid-generator", oidGeneratorClass);
        }

        // TODO load up client properties
        return properties;
    }

    private NakedObjectSpecification getSpecification(final String fullName) {
        return NakedObjectsContext.getReflector().loadSpecification(fullName);
    }

    private ReferenceData[] getDisposed() {
        Enumeration enumeration = NakedObjectsContext.getUpdateNotifer().allDisposedObjects();
        List list = new ArrayList();
        while (enumeration.hasMoreElements()) {
            NakedObject element = (NakedObject) enumeration.nextElement();
            list.add(encoder.createIdentityData(element));
        }
        return (ReferenceData[]) list.toArray(new ReferenceData[list.size()]);
    }

    private ObjectData[] getUpdates() {
        Enumeration enumeration = NakedObjectsContext.getUpdateNotifer().allChangedObjects();
        List list = new ArrayList();
        while (enumeration.hasMoreElements()) {
            NakedObject element = (NakedObject) enumeration.nextElement();
            list.add(encoder.createForUpdate(element));
        }
        return (ObjectData[]) list.toArray(new ObjectData[list.size()]);
    }

    public boolean hasInstances(final Session session, final String objectType) {
        LOG.debug("request hasInstances of " + objectType + " for " + session);
        return persistor().hasInstances(getSpecification(objectType), false);
    }

    public void init() {}

    public IdentityData oidForService(Session session, String id) {
        Object[] applicationServices = NakedObjectsContext.getObjectLoader().getServices();
        for (int i = 0; i < applicationServices.length; i++) {
            if (id.equals(ServiceUtil.id(applicationServices[i]))) {
                NakedObject adapter = NakedObjectsContext.getObjectLoader().getAdapterFor(applicationServices[i]);
                return encoder.createIdentityData(adapter);
            }
        }
        throw new NakedObjectsRemoteException("Failed to find service " + id);
    }

    private NakedObjectPersistor persistor() {
        return NakedObjectsContext.getObjectPersistor();
    }

    public Data resolveField(final Session session, final IdentityData target, final String fieldName) {
        LOG.debug("request resolveEagerly " + target + "/" + fieldName + " for " + session);

        NakedObjectSpecification spec = getSpecification(target.getType());
        NakedObjectField field = spec.getField(fieldName);
        // NakedObject object = NakedObjects.getObjectManager().getObject(target.getOid(), spec);
        NakedObject object = NakedObjectsContext.getObjectLoader().recreateAdapterForPersistent(target.getOid(), spec);
        NakedObjectsContext.getObjectPersistor().resolveField(object, field);
        return encoder.createForResolveField(object, fieldName);
    }

    public ObjectData resolveImmediately(final Session session, final IdentityData target) {
        LOG.debug("request resolveImmediately " + target + " for " + session);

        NakedObjectSpecification spec = getSpecification(target.getType());
        NakedObject object = NakedObjectsContext.getObjectPersistor().getObject(target.getOid(), spec);
        if (object.getResolveState().isResolvable(ResolveState.RESOLVING)) {
            // this is need when the object store does not load the object fully in the getObject() above
            NakedObjectsContext.getObjectPersistor().resolveImmediately(object);
        }

        return encoder.createCompletePersistentGraph(object);
    }

    /**
     * .NET property
     * 
     * @property
     */
    public void set_Encoder(final ObjectEncoder encoder) {
        setEncoder(encoder);
    }

    public ObjectData[] setAssociation(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final IdentityData associated) {
        LOG.debug("request setAssociation " + fieldIdentifier + " on " + target + " with " + associated + " for " + session);
        NakedObject inObject = getPersistentNakedObject(session, target);
        NakedObject associate = getPersistentNakedObject(session, associated);
        NakedObjectField association = (NakedObjectField) inObject.getSpecification().getField(fieldIdentifier);
        if (!association.isVisible() || association.isUsable(inObject).isVetoed()) {
            throw new IllegalRequestException("can't modify field as not visible or editable");
        }
        if (association instanceof OneToOneAssociation) {
            ((OneToOneAssociation) association).setAssociation(inObject, associate);
        } else {
            ((OneToManyAssociation) association).addElement(inObject, associate);
        }
        return getUpdates();
    }

    public void setEncoder(final ObjectEncoder objectDataFactory) {
        this.encoder = objectDataFactory;
    }

    public ObjectData[] setValue(
            final Session session,
            final String fieldIdentifier,
            final IdentityData target,
            final ValueData value) {
        Assert.assertNotNull(value);
        LOG.debug("request setValue " + fieldIdentifier + " on " + target + " with " + value + " for " + session);
        NakedObject inObject = getPersistentNakedObject(session, target);
        ValueAssociation association = (ValueAssociation) inObject.getSpecification().getField(fieldIdentifier);
        if (!association.isVisible() || association.isUsable(inObject).isVetoed()) {
            throw new IllegalRequestException("can't modify field as not visible or editable");
        }

        NakedValue fieldValue = (NakedValue) association.get(inObject);
        if (fieldValue != null) {
            // /byte[] encodedString =
            // NakedObjectsContext.getObjectLoader().createAdapterForValue(value).asEncodedString();
            fieldValue.restoreFromEncodedString(value.getEncodedValue());
        }

        NakedValue valueAdapter = NakedObjectsContext.getObjectLoader().createValueInstance(association.getSpecification());
        valueAdapter.restoreFromEncodedString(value.getEncodedValue());
        association.setValue(inObject, valueAdapter);
        return getUpdates();
    }

    public void shutdown() {}

}
// Copyright (c) Naked Objects Group Ltd.
