package org.nakedobjects.nof.reflect.security;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import org.apache.log4j.Logger;
import org.nakedobjects.noa.NakedObjectRuntimeException;
import org.nakedobjects.nof.core.conf.Configuration;
import org.nakedobjects.nof.core.context.NakedObjectsContext;
import org.nakedobjects.nof.core.util.Assert;
import org.nakedobjects.nof.reflect.peer.MemberIdentifier;
import org.nakedobjects.nof.reflect.peer.MemberIdentifierImpl;


public class FileAuthorisor implements Authorisor {
    private static final Logger LOG = Logger.getLogger(FileAuthorisor.class);
    private File whiteListFile;
    private File blackListFile;
    private Map whiteListMap = new HashMap();
    private Map blackListMap = new HashMap();
    private final boolean learn;

    private static final String AUTH_WHITE_FILE = "security.whitelist.file";
    private static final String AUTH_BLACK_FILE = "security.blacklist.file";
    private static final String AUTH_LEARN = "security.learn";

    private static final String NONE = ""; // for backward compatibility
    private static final String RO = "-ro";
    private static final String RW = "-rw";

    public FileAuthorisor() {
        learn = NakedObjectsContext.getConfiguration().getBoolean(Configuration.ROOT + AUTH_LEARN, false);
        final String whiteListFileName = NakedObjectsContext.getConfiguration().getString(Configuration.ROOT + AUTH_WHITE_FILE,
                "config/allow");
        Assert.assertTrue(whiteListFileName.length() > 0);
        whiteListFile = new File(whiteListFileName);
        if (!learn && !whiteListFile.canRead()) {
            throw new NakedObjectRuntimeException("Cannot read white list security file: " + whiteListFileName);
        }
        // black list is optional ....
        final String blackListFileName = NakedObjectsContext.getConfiguration().getString(Configuration.ROOT + AUTH_BLACK_FILE,
                "");
        if (blackListFileName.length() > 0) {
            blackListFile = new File(blackListFileName);
            if (!blackListFile.canRead()) {
                // ... but if it's there we should be able to read it
                throw new NakedObjectRuntimeException("Cannot read black list security file: " + blackListFileName);
            }
        }
    }

    private void tokenizeLine(final Map map, final String line) {
        final StringTokenizer tokens = new StringTokenizer(line, ":", false);
        Assert.assertTrue(tokens.countTokens() == 2);
        final String token1 = tokens.nextToken();
        final String token2 = tokens.nextToken();
        final MemberIdentifier memberIdentifier = memberFromString(token1.trim());
        final List roles = tokenizeRoles(token2);
        map.put(memberIdentifier.toIdentityString(MemberIdentifier.CLASS_NAME_PARMS), roles);
    }

    public void buildMap(final Map map, final File file) {
        try {
            LOG.info("loading authorisation details from " + file.getAbsolutePath());
            final FileReader fileReader = new FileReader(file);
            final BufferedReader buffReader = new BufferedReader(fileReader);
            for (String line; (line = buffReader.readLine()) != null;) {
                tokenizeLine(map, line);
            }
            fileReader.close();
            buffReader.close();
        } catch (final Exception e) {
            throw new NakedObjectRuntimeException(e);
        }
    }

    public void init() {
        if (learn) {
            return;
        }
        buildMap(whiteListMap, whiteListFile);
        if (blackListFile != null) {
            buildMap(blackListMap, blackListFile);
        }
    }

    private List tokenizeRoles(final String allRoles) {
        final List roles = new ArrayList();
        final StringTokenizer tokens = new StringTokenizer(allRoles, "|", false);
        while (tokens.hasMoreTokens()) {
            roles.add(tokens.nextToken().trim());
        }
        return roles;
    }

    private MemberIdentifier memberFromString(final String identifier) {
        return MemberIdentifierImpl.fromIdentityString(identifier);
    }

    private boolean isQualifiedMatch(final Map map, final String role, final String key, final String[] qualifiers) {
        if (map.containsKey(key)) {
            final List roles = (List) map.get(key);
            for (int i = 0; i < qualifiers.length; i++) {
                final String qualifiedRole = role + qualifiers[i];
                if (roles.contains(qualifiedRole)) {
                    return true;
                }
            }
        }
        return false;
    }

    private boolean isListed(final Map map, final String role, final MemberIdentifier member, final String[] qualifiers) {
        if (map.isEmpty()) {// quick fail
            return false;
        }
        for (int depth = MemberIdentifier.CLASS; depth <= MemberIdentifier.CLASS_NAME_PARMS; depth++) {

            if (isQualifiedMatch(map, role, member.toIdentityString(depth), qualifiers)) {
                return true;
            }
        }

        return false;
    }

    private boolean isWhiteListed(final String role, final MemberIdentifier member, final String[] qualifiers) {
        return isListed(whiteListMap, role, member, qualifiers);
    }

    private boolean isBlackListed(final String role, final MemberIdentifier member, final String[] qualifiers) {
        return isListed(blackListMap, role, member, qualifiers);
    }

    private boolean isAuthorised(final String role, final MemberIdentifier member, final String[] qualifiers) {
        if (learn) {
            return learn(role, member);
        }
        return isWhiteListed(role, member, qualifiers) && !isBlackListed(role, member, qualifiers);
    }

    public void shutdown() {
        if (learn) {
            writeMap();
        }
    }

    private void writeMap() {
        try {
            LOG.info("writing authorisation details to " + whiteListFile.getAbsolutePath());
            final FileWriter fileWriter = new FileWriter(whiteListFile);
            final BufferedWriter buffWriter = new BufferedWriter(fileWriter);
            for (int i = 0; i < whiteListMap.entrySet().size(); i++) {
                Map.Entry entry = (Map.Entry) whiteListMap.entrySet().toArray()[i];
                final StringBuffer buff = new StringBuffer();
                buff.append(entry.getKey()).append(":");
                final List roles = (List) entry.getValue();
                for (int j = 0; j < roles.size(); j++) {
                    buff.append(roles.get(j));
                    if (j < roles.size() - 1) {
                        buff.append("|");
                    }
                }
                buffWriter.write(buff.toString());
                buffWriter.newLine();
            }
            buffWriter.flush();
            buffWriter.close();
        } catch (IOException e) {
            throw new NakedObjectRuntimeException(e);
        }
    }

    private boolean learn(final String role, final MemberIdentifier member) {
        if (whiteListMap.containsKey(member.toIdentityString(MemberIdentifier.CLASS_NAME_PARMS))) {
            final List roles = (List) whiteListMap.get(member.toIdentityString(MemberIdentifier.CLASS_NAME_PARMS));
            if (!roles.contains(role)) {
                roles.add(role);
            }
        } else {
            whiteListMap.put(member.toIdentityString(MemberIdentifier.CLASS_NAME_PARMS), Arrays.asList(new String[] { role }));
        }
        return true;
    }

    public boolean isUsable(final String role, final MemberIdentifier member) {
        return isAuthorised(role, member, new String[] { NONE, RW });
    }

    public boolean isVisible(final String role, final MemberIdentifier member) {
        return isAuthorised(role, member, new String[] { NONE, RO, RW });
    }
}
